from collections.abc import Callable

import torch

from ignite.exceptions import NotComputableError
from ignite.metrics.metric import Metric, reinit__is_reduced, sync_all_reduce

__all__ = ["HitRate"]


class HitRate(Metric):
    r"""Calculates the Hit Rate at `k` for Recommendation Systems.

    The Hit Rate measures the fraction of users for which the model was able
    to predict at least one correct recommendation.
    `Hit` for each user is either 0 or 1, irrespective of how many correct recommendations
    the model was able to predict for that user.

    .. math:: \text{HR}@K = \frac{1}{N} \sum_{i=1}^{N} \mathbb{1}(\text{rank}_i \leq K)

    where :math:`\text{rank}_i` is rank of the first relevant item in the predicted
    tensor :math:`\mathbf{q}_i` that exists in the ground truth tensor :math:`\mathbf{p}_i`.

    - ``update`` must receive output of the form ``(y_pred, y)``.
    - ``y_pred`` is expected to be raw logits or probability score for each item in the catalog.
    - ``y`` is expected to be binary (only 0s and 1s) values where `1` indicates relevant item.
    - ``y_pred`` and ``y`` are only allowed shape :math:`(batch, num_items)`.
    - returns a list of HitRate ordered by the sorted values of ``top_k``.

    Args:
        top_k: a single positive integer or a list of positive integers that specifies `k` for
            calculating hitrate@top-k. If a single int is provided, it will be wrapped in a list.
            Default is 1.
        ignore_zero_hits: if True, users with no relevant items (ground truth tensor being all zeros)
            are ignored in computation of HitRate. if set False, such users are counted as a miss.
            By default, True.
        output_transform: a callable that is used to transform the
            :class:`~ignite.engine.engine.Engine`'s ``process_function``'s output into the
            form expected by the metric.
            The output is expected to be a tuple `(prediction, target)`
            where `prediction` and `target` are tensors
            of shape ``(batch, num_items)``.
        device: specifies which device updates are accumulated on. Setting the
            metric's device to be the same as your ``update`` arguments ensures the ``update`` method is
            non-blocking. By default, CPU.
        skip_unrolling: specifies whether input should be unrolled or not before being
            processed. Should be true for multi-output models..

    Examples:
        To use with ``Engine`` and ``process_function``, simply attach the metric instance to the engine.
        The output of the engine's ``process_function`` needs to be in the format of
        ``(y_pred, y)``. If not, ``output_tranform`` can be added
        to the metric to transform the output into the form expected by the metric.

        For more information on how metric works with :class:`~ignite.engine.engine.Engine`, visit :ref:`attach-engine`.

        .. include:: defaults.rst
            :start-after: :orphan:

        ignore_zero_hits=True case

        .. testcode:: 1

            metric = HitRate(top_k=[1, 2, 3, 4])
            metric.attach(default_evaluator,"hit_rate")
            y_pred=torch.Tensor([
                [4.0, 2.0, 3.0, 1.0],
                [1.0, 2.0, 3.0, 4.0]
            ])
            y_true=torch.Tensor([
                [0.0, 0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0, 0.0]
            ])
            state = default_evaluator.run([(y_pred, y_true)])
            print(state.metrics["hit_rate"])

        .. testoutput:: 1

            [0.0, 1.0, 1.0, 1.0]

        ignore_zero_hits=False case

        .. testcode:: 2

            metric = HitRate(top_k=[1, 2, 3, 4], ignore_zero_hits=False)
            metric.attach(default_evaluator,"hit_rate")
            y_pred=torch.Tensor([
                [4.0, 2.0, 3.0, 1.0],
                [1.0, 2.0, 3.0, 4.0]
            ])
            y_true=torch.Tensor([
                [0.0, 0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0, 0.0]
            ])
            state = default_evaluator.run([(y_pred, y_true)])
            print(state.metrics["hit_rate"])

        .. testoutput:: 2

            [0.0, 0.5, 0.5, 0.5]

        int top_k case

        .. testcode:: 3

            metric = HitRate(top_k=2)
            metric.attach(default_evaluator, "hit_rate")
            y_pred = torch.Tensor([
                [4.0, 2.0, 3.0, 1.0],
            ])
            y_true = torch.Tensor([
                [0.0, 0.0, 1.0, 0.0],
            ])
            state = default_evaluator.run([(y_pred, y_true)])
            print(state.metrics["hit_rate"])

        .. testoutput:: 3

            [1.0]

    .. versionadded:: 0.5.4
    .. versionchanged:: 0.5.4
        `top_k` now accepts a single positive integer in addition to a list of integers.
    """

    required_output_keys = ("y_pred", "y")
    _state_dict_all_req_keys = ("_hits_per_k", "_num_examples")

    def __init__(
        self,
        top_k: list[int] | int = 1,
        ignore_zero_hits: bool = True,
        output_transform: Callable = lambda x: x,
        device: str | torch.device = torch.device("cpu"),
        skip_unrolling: bool = False,
    ):
        if not isinstance(top_k, (int, list)):
            raise ValueError("top_k must be either int or a list[int]")

        top_k = [top_k] if isinstance(top_k, int) else top_k

        if len(top_k) == 0:
            raise ValueError("top_k must have at least one positive value")
        if any(k <= 0 for k in top_k):
            raise ValueError("top_k must be list of positive integers only.")

        self.top_k = sorted(top_k)
        self.ignore_zero_hits = ignore_zero_hits
        super().__init__(output_transform, device=device, skip_unrolling=skip_unrolling)

    @reinit__is_reduced
    def reset(self) -> None:
        self._hits_per_k = torch.zeros(len(self.top_k), device=self._device)
        self._num_examples = 0

    @reinit__is_reduced
    def update(self, output: tuple[torch.Tensor, torch.Tensor]) -> None:
        if len(output) != 2:
            raise ValueError(f"output should be in format `(y_pred,y)` but got tuple of {len(output)} tensors.")

        y_pred, y = output
        if y_pred.shape != y.shape:
            raise ValueError(f"y_pred and y must be in the same shape, got {y_pred.shape} != {y.shape}.")

        if self.ignore_zero_hits:
            valid_mask = torch.any(y > 0, dim=-1)
            y_pred = y_pred[valid_mask]
            y = y[valid_mask]

        if y.shape[0] == 0:
            return

        max_k = self.top_k[-1]
        _, indices = torch.topk(y_pred, k=max_k, dim=-1)

        hits_at_max_k = torch.gather(y, dim=-1, index=indices)

        for i, k in enumerate(self.top_k):
            hit_mask = torch.any(hits_at_max_k[:, :k] > 0, dim=-1)
            self._hits_per_k[i] += torch.sum(hit_mask).to(self._device)

        self._num_examples += y.shape[0]

    @sync_all_reduce("_hits_per_k", "_num_examples")
    def compute(self) -> list[float]:
        if self._num_examples == 0:
            raise NotComputableError("HitRate must have at least one example.")

        rates = (self._hits_per_k / self._num_examples).tolist()
        return rates



from ignite.metrics.fairness.accuracy_difference import SubgroupAccuracyDifference
from ignite.metrics.fairness.base import SubgroupDifference, SubgroupMetric
from ignite.metrics.fairness.demographic_parity import DemographicParityDifference, SelectionRate



print("import: 'ignite'")
import ignite

print("import: 'ignite.engine'")
import ignite.engine

print("import: 'ignite.handlers'")
import ignite.handlers

print("import: 'ignite.metrics'")
import ignite.metrics

print("import: 'ignite.contrib'")
import ignite.contrib




from collections.abc import Callable

import torch

from ignite.metrics.epoch_metric import EpochMetric


def matthews_corrcoef_compute_fn(y_preds: torch.Tensor, y_targets: torch.Tensor) -> float:
    from sklearn.metrics import matthews_corrcoef

    if y_preds.ndim == 2 and y_targets.ndim == 2:
        y_preds = torch.argmax(y_preds, dim=1)
        y_targets = torch.argmax(y_targets, dim=1)
    elif y_preds.ndim == 2 and y_targets.ndim == 1:
        y_preds = torch.argmax(y_preds, dim=1)
    elif y_preds.ndim == 1 and y_targets.ndim == 2:
        raise ValueError(
            "Incoherent types between input y_pred and stored predictions: y_pred is 1D while y_target is 2D"
        )

    y_true = y_targets.cpu().numpy()
    y_pred = y_preds.cpu().numpy()
    return matthews_corrcoef(y_true, y_pred)


class MatthewsCorrCoef(EpochMetric):
    """
    Compute the Matthews correlation coefficient (MCC).

    The Matthews correlation coefficient is used in machine learning as a measure of the quality of binary and multiclass classifications.
    It takes into account true and false positives and negatives and is generally regarded as a balanced measure which can be used even if the classes are of very different sizes.
    The MCC is in essence a correlation coefficient value between -1 and +1. A coefficient of +1 represents a perfect prediction, 0 an average random prediction and -1 an inverse prediction.

    This metric is suitable for both binary and multiclass classification.
    In the binary case, it is calculated using the entries of the confusion matrix, whereas for multiclass tasks, it is computed as a generalized correlation coefficient.

    In case of multiclass classification with shape (N, C) for y_pred and (N, C) for y, the predicted class is determined by the argmax of y_pred and y.
    In case of multiclass classification with shape (N, C) for y_pred and (N,) for y, the predicted class is determined by the argmax of y_pred and the true class is determined by the value in y.

        Args:
            output_transform: a callable that is used to transform the
                :class:`~ignite.engine.engine.Engine`'s ``process_function``'s output into the
                form expected by the metric. This can be useful if, for example, you have a multi-output model and
                you want to compute the metric with respect to one of the outputs.
                By default, this metric requires the output as ``(x, y)``.
            device: specifies which device updates are accumulated on. Setting the
                metric's device to be the same as your ``update`` arguments ensures the ``update`` method is
                non-blocking. By default, CPU.
            check_compute_fn: if True, compute_fn is run on the first batch of data to ensure there are no issues.
                If issues exist, user is warned that there might be an issue with the compute_fn. Default, True.
            skip_unrolling: specifies whether output should be unrolled before being fed to update method. Should be
                true for multi-output model, for example, if ``y_pred`` contains multi-output as ``(y_pred_a, y_pred_b)``
                Alternatively, ``output_transform`` can be used to handle this.

            Examples:

            .. include:: defaults.rst
                :start-after: :orphan:


            .. testcode::

                y_pred = torch.tensor([+1, +1, +1, -1])
                y_true = torch.tensor([+1, -1, +1, +1])

                matthews_corrcoef = MatthewsCorrCoef()
                matthews_corrcoef.attach(default_evaluator, 'mcc')
                state = default_evaluator.run([[y_pred, y_true]])
                print(state.metrics['mcc'])

            .. testoutput::

                -0.33...

        .. versionadded:: 0.5.4

    """

    def __init__(
        self,
        output_transform: Callable = lambda x: x,
        check_compute_fn: bool = False,
        device: str | torch.device = torch.device("cpu"),
        skip_unrolling: bool = False,
    ):
        try:
            from sklearn.metrics import matthews_corrcoef  # noqa: F401
        except ImportError:
            raise ModuleNotFoundError("This metric module requires scikit-learn to be installed.")

        super().__init__(
            matthews_corrcoef_compute_fn,
            output_transform=output_transform,
            check_compute_fn=check_compute_fn,
            device=device,
            skip_unrolling=skip_unrolling,
        )



import torch
from collections.abc import Callable, Sequence
from ignite.metrics.accuracy import Accuracy
from ignite.metrics.fairness.base import SubgroupDifference

__all__ = ["SubgroupAccuracyDifference"]


class SubgroupAccuracyDifference(SubgroupDifference):
    r"""Calculates the Subgroup Accuracy Difference.

    This metric computes the accuracy for each unique subgroup in the dataset and returns
    the maximum difference in accuracy between any two subgroups. It is a strict measure
    of how disparate the performance of a model is across different categorical segments.

    This metric is referred to as *Overall Accuracy Equality* in the fairness literature.

    - ``update`` must receive output of the form ``(y_pred, y, group_labels)`` or
      ``{'y_pred': y_pred, 'y': y, 'group_labels': group_labels}``.
    - `y_pred` must be in the following shape (batch_size, num_categories, ...) or (batch_size, ...).
    - `y` must be in the following shape (batch_size, ...).
    - `group_labels` must be a 1D tensor of shape (batch_size,) containing discrete labels.

    Args:
        groups: a sequence of unique group identifiers.
        is_multilabel: if True, multilabel accuracy is calculated. By default, False.
        output_transform: a callable that is used to transform the
            :class:`~ignite.engine.engine.Engine`'s ``process_function``'s output into the
            form expected by the metric.
        device: specifies which device updates are accumulated on. Setting the
            metric's device to be the same as your ``update`` arguments ensures the ``update`` method is
            non-blocking. By default, CPU.

    Examples:
        To use with ``Engine`` and ``process_function``, simply attach the metric instance to the engine.
        The output of the engine's ``process_function`` needs to be in the format of
        ``(y_pred, y, group_labels)``.

        .. include:: defaults.rst
            :start-after: :orphan:

        .. testcode::

            metric = SubgroupAccuracyDifference(groups=[0, 1])
            metric.attach(default_evaluator, 'subgroup_acc_diff')

            # Predictions for 4 items:
            # Items 1 and 3 are predicted as class 0 (index 0 has highest prob)
            # Items 2 and 4 are predicted as class 1 (index 1 has highest prob)
            y_pred = torch.tensor([[0.9, 0.1], [0.1, 0.9], [0.8, 0.2], [0.2, 0.8]])

            # Targets
            y_true = torch.tensor([0, 1, 1, 0])

            # Subgroups (e.g., 0=Demographic A, 1=Demographic B)
            group_labels = torch.tensor([0, 0, 1, 1])

            # Subgroup 0: 2 correct predictions, accuracy = 100%
            # Subgroup 1: 0 correct predictions, accuracy = 0%

            state = default_evaluator.run([[y_pred, y_true, group_labels]])
            print(state.metrics['subgroup_acc_diff'])

        .. testoutput::

            1.0

    .. versionadded:: 0.5.4

    References:
        - Verma & Rubin, `Fairness Definitions Explained
          <https://fairware.cs.umass.edu/papers/Verma.pdf>`_, 2018.
    """

    def __init__(
        self,
        groups: Sequence[int],
        is_multilabel: bool = False,
        output_transform: Callable = lambda x: x,
        device: torch.device | str = torch.device("cpu"),
    ) -> None:
        acc = Accuracy(is_multilabel=is_multilabel, device=device)
        super().__init__(base_metric=acc, groups=groups, output_transform=output_transform, device=device)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections.abc import Callable
from numbers import Number
from pathlib import Path
from typing import Any, cast

import torch
from packaging.version import Version

_torch_version_gt_112 = Version(torch.__version__) > Version("1.12.0")


class ComputationModel(metaclass=ABCMeta):
    """Base class for distributed computation models and defines interface methods.
    This class is public and should be used for other custom derived distributed models.
    """

    # this is an additional local rank storage used when idist is setup from existing native torch dist context
    _ext_local_rank: int | None = None

    def __init__(self) -> None:
        self._backend: str | None = None
        self._nproc_per_node: int | None = None
        self._nnodes: int | None = None
        self._node: int | None = None

    def _setup_attrs(self) -> None:
        if self._nproc_per_node is None:
            self._nproc_per_node = self._compute_nproc_per_node() if self.get_world_size() > 1 else 1
        if self._nnodes is None:
            self._nnodes = self.get_world_size() // self._nproc_per_node
        if self._node is None:
            self._node = self.get_rank() // self._nproc_per_node

    @abstractmethod
    def _compute_nproc_per_node(self) -> int:
        pass

    @abstractmethod
    def get_local_rank(self) -> int:
        pass

    @abstractmethod
    def get_rank(self) -> int:
        pass

    @abstractmethod
    def get_world_size(self) -> int:
        pass

    @abstractmethod
    def get_nproc_per_node(self) -> int:
        pass

    @abstractmethod
    def get_nnodes(self) -> int:
        pass

    @abstractmethod
    def get_node_rank(self) -> int:
        pass

    @abstractmethod
    def device(self) -> torch.device:
        pass

    @abstractmethod
    def backend(self) -> str | None:
        pass

    @abstractmethod
    def finalize(self) -> None:
        pass

    @staticmethod
    @abstractmethod
    def create_from_context() -> ComputationModel | None:
        pass

    @staticmethod
    @abstractmethod
    def create_from_backend(backend: str, **kwargs: Any) -> ComputationModel:
        pass

    @staticmethod
    @abstractmethod
    def spawn(*args: Any, **kwargs: Any) -> None:
        pass

    _collective_op_dtype: Any = None

    @staticmethod
    def _encode_str(x: str, device: torch.device, size: int) -> torch.Tensor:
        name = torch.tensor(bytearray(x, "utf-8")).to(device)
        padded_x = torch.zeros(size + 1, device=device, dtype=torch.long)
        padded_x[: len(name)] = name
        padded_x[-1] = len(name)
        # output is tensor of shape (1, size + 1)
        return padded_x.unsqueeze(0)

    def _get_max_length(self, x: str, device: torch.device) -> int:
        size = torch.tensor([len(x)], device=device)
        size = self._do_all_reduce(size, op="MAX")
        return cast(int, size.item())

    @staticmethod
    def _encode_input_data(x: torch.Tensor | float | str | Path | None, is_src: bool) -> list[int]:
        encoded_msg = [-1] * 512
        if not is_src:
            # Discard input type if not source
            return encoded_msg

        if isinstance(x, torch.Tensor):
            shape = x.shape
            dtype = str(x.dtype)
            msg = [0, len(shape), *shape, len(dtype), *list(bytearray(dtype, "utf-8"))]
            encoded_msg[: len(msg)] = msg
        elif isinstance(x, Number):
            encoded_msg[0] = 1
        elif isinstance(x, str):
            encoded_msg[0] = 2
        elif isinstance(x, Path):
            # Keep a distinct tag so non-source ranks in safe mode can reconstruct
            # Path placeholders and return Path values rather than strings.
            encoded_msg[0] = 3
        return encoded_msg

    @staticmethod
    def _decode_as_placeholder(encoded_msg: list[int], device: torch.device) -> torch.Tensor | float | str | Path:
        if encoded_msg[0] == 0:
            len_shape = encoded_msg[1]
            le = 2 + len_shape
            shape = encoded_msg[2:le] if len_shape > 0 else []
            len_dtype = encoded_msg[le]
            dtype_str = bytearray(encoded_msg[le + 1 : le + 1 + len_dtype]).decode("utf-8")
            dtype = eval(dtype_str)
            return torch.empty(shape, device=device, dtype=dtype)
        elif encoded_msg[0] == 1:
            return 0.0
        elif encoded_msg[0] == 2:
            return ""
        elif encoded_msg[0] == 3:
            return Path("")
        else:
            raise RuntimeError(f"Internal error: unhandled dtype {encoded_msg[0]}, encoded_msg={encoded_msg}")

    def _setup_placeholder(
        self, x: torch.Tensor | float | str | Path | None, device: torch.device, is_src: bool
    ) -> torch.Tensor | float | str | Path:
        encoded_msg_per_rank = self._encode_input_data(x, is_src)
        encoded_msg_all_ranks = self._do_all_reduce(torch.tensor(encoded_msg_per_rank, device=device), op="MAX")

        if is_src:
            if x is None:
                raise RuntimeError("Internal error, x is None. Please, file an issue if you encounter this error.")
            return x

        encoded_msg = encoded_msg_all_ranks.cpu().tolist()
        return self._decode_as_placeholder(encoded_msg, device)

    @staticmethod
    def _decode_str(xs: torch.Tensor) -> list[str]:
        # xs.shape = (n, size + 1), e.g. (world_size, size + 1)
        out = [bytearray(x[: x[-1]].tolist()).decode("utf-8") for x in xs]
        return out

    def _apply_op(
        self, tensor: torch.Tensor, device: torch.device, fn: Callable, *args: Any, **kwargs: Any
    ) -> torch.Tensor:
        out_dtype = None
        tensor_device = None

        # check if the tensor is at specified device
        if tensor.device != device:
            tensor_device = tensor.device
            tensor = tensor.to(device)

        if self._collective_op_dtype is not None:
            # cast to _collective_op_dtype if current type is not floatX
            if tensor.dtype not in (torch.float32, torch.float64):
                out_dtype = tensor.dtype
                tensor = tensor.to(self._collective_op_dtype)

        tensor = fn(tensor, *args, **kwargs)

        if out_dtype is not None and tensor_device is not None:
            return tensor.to(dtype=out_dtype, device=tensor_device)
        if out_dtype is not None:
            return tensor.to(dtype=out_dtype)
        if tensor_device is not None:
            return tensor.to(device=tensor_device)
        return tensor

    def _collective_op(
        self, tensor: torch.Tensor | Number | str, fn: Callable, *args: Any, **kwargs: Any
    ) -> torch.Tensor | float | list[float] | list[str]:
        tensor_to_number = tensor_to_str = False
        device = self.device()
        if isinstance(tensor, (Number, float)):
            tensor_to_number = True
            dtype = self._collective_op_dtype
            if dtype is None and isinstance(tensor, float):
                dtype = torch.double
            tensor = torch.tensor(tensor, device=device, dtype=dtype)
        elif isinstance(tensor, str):
            tensor_to_str = True
            max_length = self._get_max_length(tensor, device)
            tensor = self._encode_str(tensor, device, size=max_length)

        tensor = self._apply_op(tensor, device, fn, *args, **kwargs)

        if tensor_to_number:
            return tensor.tolist()
        elif tensor_to_str:
            return self._decode_str(tensor)
        return tensor

    def all_reduce(
        self, tensor: torch.Tensor | float, op: str = "sum", group: Any | None = None
    ) -> torch.Tensor | float:
        if not isinstance(tensor, (torch.Tensor, Number)):
            raise TypeError(f"Unhandled input type {type(tensor)}")

        return cast(torch.Tensor | float, self._collective_op(tensor, self._do_all_reduce, op, group=group))

    def all_gather(
        self, tensor: torch.Tensor | float | str | Any, group: Any | None = None
    ) -> torch.Tensor | float | list[float] | list[str] | list[Any]:
        if not isinstance(tensor, (torch.Tensor, Number, str)):
            return self._do_all_gather_object(tensor, group=group)

        return self._collective_op(tensor, self._do_all_gather, group=group)

    def new_group(self, ranks: list[int], **kwargs: Any) -> Any:
        if isinstance(ranks, list) and all(isinstance(item, int) for item in ranks):
            return self._do_new_group(ranks, **kwargs)
        else:
            raise ValueError("Argument ranks should be list of int")

    def broadcast(
        self, tensor: torch.Tensor | float | str | Path | None, src: int = 0, safe_mode: bool = False
    ) -> torch.Tensor | float | str | Path:
        if not (isinstance(tensor, (torch.Tensor, Number, str, Path)) or tensor is None):
            raise TypeError(f"Unhandled input type {type(tensor)}")

        rank = self.get_rank()
        if tensor is None:
            if rank == src:
                raise ValueError("Source data can not be None")
            elif not safe_mode:
                raise ValueError("Argument safe_mode should be True if None is passed for non src rank")

        device = self.device()
        tensor_to_number = tensor_to_str = tensor_to_path = False

        if safe_mode:
            tensor = self._setup_placeholder(tensor, device, rank == src)

        if tensor is None:
            raise RuntimeError("Internal error, tensor is None. Please, file an issue if you encounter this error.")

        if isinstance(tensor, (Number, float)):  # have to use Number and float to satisfy mypy
            tensor_to_number = True
            if rank != src:
                tensor = torch.empty(1, device=device, dtype=torch.float)
            else:
                tensor = torch.tensor([tensor], device=device, dtype=torch.float)
        elif isinstance(tensor, (str, Path)):
            tensor_to_str = isinstance(tensor, str)
            tensor_to_path = isinstance(tensor, Path)
            tensor_str = str(tensor)
            max_length = self._get_max_length(tensor_str, device)
            if rank != src:
                tensor = torch.empty(1, max_length + 1, device=device, dtype=torch.long)
            else:
                tensor = self._encode_str(tensor_str, device, size=max_length)

        tensor = self._apply_op(tensor, device, self._do_broadcast, src)

        if tensor_to_number:
            return tensor.item()
        if tensor_to_str or tensor_to_path:
            list_str = self._decode_str(tensor)
            result = list_str[0]
            return Path(result) if tensor_to_path else result
        return tensor

    @abstractmethod
    def _do_all_reduce(self, tensor: torch.Tensor, op: str = "SUM", group: Any | None = None) -> torch.Tensor:
        pass

    @abstractmethod
    def _do_all_gather(self, tensor: torch.Tensor, group: Any | None = None) -> torch.Tensor:
        pass

    @abstractmethod
    def _do_all_gather_object(self, tensor: Any, group: Any | None = None) -> list[Any]:
        pass

    @abstractmethod
    def _do_broadcast(self, tensor: torch.Tensor, src: int) -> torch.Tensor:
        pass

    @abstractmethod
    def barrier(self) -> None:
        pass

    @abstractmethod
    def _do_new_group(self, ranks: list[int], **kwargs: Any) -> Any:
        pass

    @abstractmethod
    def _rank_not_in_group(self, group: Any) -> bool:
        pass


class _SerialModel(ComputationModel):
    """Private class defines non-distributed computation model for code compatibility with other distributed models."""

    name = "serial"
    available_backends = ()

    def __init__(self, _backend: str | None = None, **kwargs: Any) -> None:
        super().__init__()

    def get_local_rank(self) -> int:
        return 0

    def get_rank(self) -> int:
        return 0

    def get_world_size(self) -> int:
        return 1

    def get_nproc_per_node(self) -> int:
        return 1

    def get_nnodes(self) -> int:
        return 1

    def get_node_rank(self) -> int:
        return 0

    def device(self) -> torch.device:
        if torch.cuda.is_available():
            return torch.device("cuda")
        if _torch_version_gt_112 and torch.backends.mps.is_available():
            return torch.device("mps")
        return torch.device("cpu")

    def backend(self) -> str | None:
        return None

    def finalize(self) -> None:
        pass

    def _compute_nproc_per_node(self) -> int:
        return 1

    @staticmethod
    def create_from_context() -> _SerialModel:
        return _SerialModel()

    @staticmethod
    def create_from_backend(backend: str | None = None, **kwargs: Any) -> _SerialModel:
        return _SerialModel()

    @staticmethod
    def spawn(*args: Any, **kwargs: Any) -> None:
        raise NotImplementedError("Serial computation model does not implement spawn method")

    def all_reduce(
        self, tensor: torch.Tensor | float, op: str = "SUM", group: Any | None = None
    ) -> torch.Tensor | float:
        return tensor

    def all_gather(
        self, tensor: torch.Tensor | float | str | Any, group: Any | None = None
    ) -> torch.Tensor | float | list[float] | list[str] | list[Any]:
        if isinstance(tensor, torch.Tensor):
            return tensor
        return cast(list[float] | list[str] | list[Any], [tensor])

    def broadcast(
        self, tensor: torch.Tensor | float | str | Path | None, src: int = 0, safe_mode: bool = False
    ) -> torch.Tensor | float | str | Path:
        if tensor is None:
            raise ValueError("Argument tensor should not be None")
        return tensor

    def _do_all_reduce(self, tensor: torch.Tensor, op: str = "SUM", group: Any | None = None) -> torch.Tensor:
        return tensor

    def _do_all_gather(self, tensor: torch.Tensor, group: Any | None = None) -> torch.Tensor:
        return tensor

    def _do_all_gather_object(self, tensor: Any, group: Any | None = None) -> Any:
        return tensor

    def _do_new_group(self, ranks: list[int], **kwargs: Any) -> Any:
        return ranks

    def _do_broadcast(self, tensor: torch.Tensor, src: int) -> torch.Tensor:
        return tensor

    def barrier(self) -> None:
        pass

    def new_group(self, ranks: list[int], **kwargs: Any) -> Any:
        if isinstance(ranks, list) and all(isinstance(item, int) for item in ranks):
            return self._do_new_group(ranks, **kwargs)
        else:
            raise ValueError("Argument ranks should be list of int")

    def _rank_not_in_group(self, group: Any) -> bool:
        return False



from typing import Callable

import torch

from ignite.exceptions import NotComputableError
from ignite.metrics.metric import Metric, reinit__is_reduced, sync_all_reduce

__all__ = ["NDCG"]


def _get_ranked_items(scores: torch.Tensor, items: torch.Tensor, k: int) -> torch.Tensor:
    """Get top-k items ranked by scores."""
    indices = torch.argsort(scores, dim=-1, descending=True, stable=True)[:, :k]
    return torch.gather(items, dim=-1, index=indices)


class NDCG(Metric):
    r"""Calculates the Normalized Discounted Cumulative Gain (NDCG) at `k` for Recommendation Systems.

    For a step-by-step guide on how to use this metric, please refer to the
    `NDCG Tutorial <https://github.com/pytorch-ignite/examples/tree/main/tutorials/intermediate/ndcg-metric-tutorial.ipynb>`

    NDCG measures the quality of ranking by considering both the relevance of items and their
    positions in the ranked list. It compares the achieved DCG against the ideal DCG (IDCG)
    obtained by sorting items by their true relevance.

    .. math::
        \text{NDCG}@K = \frac{\text{DCG}@K}{\text{IDCG}@K}

    where:

    .. math::
        \text{DCG}@K = \sum_{i=1}^{K} \frac{2^{\text{rel}_i} - 1}{\log_2(i + 1)}

    and :math:`\text{rel}_i` is the relevance score of the item at position :math:`i` in the
    ranked list (1-indexed).

    - ``update`` must receive output of the form ``(y_pred, y)``.
    - ``y_pred`` is expected to be raw logits or probability score for each item in the catalog.
    - ``y`` is expected to contain relevance scores (can be binary or graded).

      Relevance Types:
      - **Binary relevance**: Labels are either 0 (not relevant) or 1 (relevant)
      - **Graded relevance**: Labels can have multiple levels (e.g., 0-4 scale)

        Common graded scales:
        - 0: Not relevant
        - 1: Marginally relevant
        - 2: Relevant
        - 3: Highly relevant
        - 4: Perfectly relevant

      The NDCG formula handles both types through the gain function: 2^relevance - 1.
      Higher relevance scores contribute more to the metric.

    - ``y_pred`` and ``y`` are only allowed shape :math:`(batch, num\_items)`.
    - returns a list of NDCG ordered by the sorted values of ``top_k``.

    Args:
        top_k: a list of sorted positive integers that specifies `k` for calculating NDCG@top-k.
        ignore_zero_hits: if True, users with no relevant items (ground truth tensor being all zeros)
            are ignored in computation of NDCG. If set False, such users are counted with NDCG of 0.
            By default, True.
        relevance_threshold: minimum label value to be considered relevant. Defaults to ``1``,
            which handles standard binary labels and graded relevance scales (e.g. TREC-style
            0-4) by treating any label >= 1 as relevant. Items below this threshold contribute
            0 to DCG/IDCG calculations.
        gain_function (str): Gain function for relevance scores. Options:
            - ``'exp_rank'``: 2^relevance - 1 (emphasizes high relevance, default)
            - ``'linear_rank'``: relevance (simpler, linear scale)
            Defaults to ``'exp_rank'``.
        output_transform: a callable that is used to transform the
            :class:`~ignite.engine.engine.Engine`'s ``process_function``'s output into the
            form expected by the metric.
            The output is expected to be a tuple `(prediction, target)`
            where `prediction` and `target` are tensors
            of shape ``(batch, num_items)``.
        device: specifies which device updates are accumulated on. Setting the
            metric's device to be the same as your ``update`` arguments ensures the ``update`` method is
            non-blocking. By default, CPU.
        skip_unrolling: specifies whether input should be unrolled or not before being
            processed. Should be true for multi-output models..

    Examples:
        To use with ``Engine`` and ``process_function``, simply attach the metric instance to the engine.
        The output of the engine's ``process_function`` needs to be in the format of
        ``(y_pred, y)``. If not, ``output_tranform`` can be added
        to the metric to transform the output into the form expected by the metric.

        For more information on how metric works with :class:`~ignite.engine.engine.Engine`, visit :ref:`attach-engine`.

        .. include:: defaults.rst
            :start-after: :orphan:

        ignore_zero_hits=True case

        .. testcode:: 1

            metric = NDCG(top_k=[1, 2, 3, 4])
            metric.attach(default_evaluator, "ndcg")
            y_pred=torch.Tensor([
                [4.0, 2.0, 3.0, 1.0],
                [1.0, 2.0, 3.0, 4.0]
            ])
            y_true=torch.Tensor([
                [0.0, 0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0, 0.0]
            ])
            state = default_evaluator.run([(y_pred, y_true)])
            print(state.metrics["ndcg"])

        .. testoutput:: 1

            [0.0, 0.38..., 0.38..., 0.65...]

        ignore_zero_hits=False case

        .. testcode:: 2

            metric = NDCG(top_k=[1, 2, 3, 4], ignore_zero_hits=False)
            metric.attach(default_evaluator, "ndcg")
            y_pred=torch.Tensor([
                [4.0, 2.0, 3.0, 1.0],
                [1.0, 2.0, 3.0, 4.0]
            ])
            y_true=torch.Tensor([
                [0.0, 0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0, 0.0]
            ])
            state = default_evaluator.run([(y_pred, y_true)])
            print(state.metrics["ndcg"])

        .. testoutput:: 2

            [0.0, 0.19..., 0.19..., 0.32...]

    .. versionadded:: 0.6.0
    """

    required_output_keys = ("y_pred", "y")
    _state_dict_all_req_keys = ("_sum_ndcg_per_k", "_num_examples")

    def __init__(
        self,
        top_k: list[int],
        ignore_zero_hits: bool = True,
        relevance_threshold: float = 1.0,
        gain_function: str = "exp_rank",
        output_transform: Callable = lambda x: x,
        device: str | torch.device = torch.device("cpu"),
        skip_unrolling: bool = False,
    ):
        if any(k <= 0 for k in top_k):
            raise ValueError(f"top_k must be list of positive integers only, but given {top_k}")

        if gain_function not in ["exp_rank", "linear_rank"]:
            raise ValueError(f"gain_function must be either 'exp_rank' or 'linear_rank', but given {gain_function}")

        self.top_k = sorted(top_k)
        self.ignore_zero_hits = ignore_zero_hits
        self.relevance_threshold = relevance_threshold
        self.gain_function = gain_function
        super(NDCG, self).__init__(output_transform, device=device, skip_unrolling=skip_unrolling)

    @reinit__is_reduced
    def reset(self) -> None:
        self._sum_ndcg_per_k = torch.zeros(len(self.top_k), device=self._device)
        self._num_examples = 0

    def _compute_dcg(self, relevance_scores: torch.Tensor, k: int) -> torch.Tensor:
        """Compute DCG@k for a batch of relevance scores."""
        actual_k = min(k, relevance_scores.shape[1])

        positions = torch.arange(1, actual_k + 1, dtype=torch.float32, device=relevance_scores.device)
        discounts = 1.0 / torch.log2(positions + 1)

        topk_relevance = relevance_scores[:, :actual_k]
        if self.gain_function == "exp_rank":
            gains = torch.pow(2.0, topk_relevance) - 1
        else:
            gains = topk_relevance

        dcg = (gains * discounts).sum(dim=-1)
        return dcg

    @reinit__is_reduced
    def update(self, output: tuple[torch.Tensor, torch.Tensor]) -> None:
        if len(output) != 2:
            raise ValueError(f"output should be in format `(y_pred,y)` but got tuple of {len(output)} tensors.")

        y_pred, y = output
        if y_pred.shape != y.shape:
            raise ValueError(f"y_pred and y must be in the same shape, got {y_pred.shape} != {y.shape}.")

        if self.ignore_zero_hits:
            valid_mask = torch.any(y >= self.relevance_threshold, dim=-1)
            y_pred = y_pred[valid_mask]
            y = y[valid_mask]

        if y.shape[0] == 0:
            return

        y_for_dcg = torch.where(y >= self.relevance_threshold, y, 0)

        max_k = self.top_k[-1]
        ranked_relevance = _get_ranked_items(y_pred, y_for_dcg, max_k)
        # Compute ideal ranking by sorting true relevance scores (descending).
        # This aligns with standard IDCG computation in reference libraries:
        # ranx: https://github.com/AmenRa/ranx/blob/master/ranx/metrics/ndcg.py#L52
        # catalyst: https://github.com/catalyst-team/catalyst/blob/master/catalyst/metrics/functional/_ndcg.py#L197

        ideal_relevance = torch.sort(y_for_dcg, dim=-1, descending=True, stable=True)[0][:, :max_k]

        for i, k in enumerate(self.top_k):
            dcg_k = self._compute_dcg(ranked_relevance, k)
            idcg_k = self._compute_dcg(ideal_relevance, k)

            ndcg_k = torch.where(
                idcg_k > 0,
                dcg_k / idcg_k,
                torch.zeros_like(dcg_k),
            )

            self._sum_ndcg_per_k[i] += ndcg_k.sum().to(self._device)

        self._num_examples += y.shape[0]

    @sync_all_reduce("_sum_ndcg_per_k", "_num_examples")
    def compute(self) -> list[float]:
        if self._num_examples == 0:
            raise NotComputableError("NDCG must have at least one example.")

        rates = (self._sum_ndcg_per_k / self._num_examples).tolist()
        return rates



from collections import OrderedDict
from collections.abc import Callable, Mapping
from typing import Any, cast, Literal
import warnings

from ignite.base import Serializable, ResettableHandler
from ignite.engine import Engine, Events
from ignite.utils import setup_logger

__all__ = ["EarlyStopping"]


class EarlyStopping(Serializable, ResettableHandler):
    """EarlyStopping handler can be used to stop the training if no improvement after a given number of events.

    Args:
        patience: Number of events to wait if no improvement and then stop the training.
        score_function: It should be a function taking a single argument, an
            :class:`~ignite.engine.engine.Engine` object, and return a score ``float``.
            An improvement is considered if the score is higher (for ``mode='max'``)
            or lower (for ``mode='min'``).
        trainer: Trainer engine to stop the run if no improvement.
        threshold: A minimum change in the score to qualify as an improvement.
            For ``mode='max'``, it is a minimum increase; for ``mode='min'``, it is a
            minimum decrease. An improvement is only considered if the change
            exceeds the threshold determined by ``threshold`` and ``threshold_mode``.
        cumulative: If True, ``threshold`` defines the change since the last
            ``patience`` reset, otherwise it defines the change after the last
            event. Default value is False.
        threshold_mode: Determines whether ``threshold`` is an absolute change
            or a relative change.

            - In ``'abs'`` mode:

              - For ``mode='max'``: improvement if ``score > best_score + threshold``
              - For ``mode='min'``: improvement if ``score < best_score - threshold``

            - In ``'rel'`` mode:

              - For ``mode='max'``: improvement if ``score > best_score * (1 + threshold)``
              - For ``mode='min'``: improvement if ``score < best_score * (1 - threshold)``

            Possible values are ``"abs"`` and ``"rel"``. Default value is ``"abs"``.
        mode: Whether to maximize (``'max'``) or minimize (``'min'``) the score.
            Default is ``'max'``.

    Examples:
        .. code-block:: python

            from ignite.engine import Engine, Events
            from ignite.handlers import EarlyStopping

            def score_function(engine):
                val_loss = engine.state.metrics["nll"]
                return -val_loss

            handler = EarlyStopping(
                patience=10,
                score_function=score_function,
                trainer=trainer,
            )

            # Note: the handler is attached to an *Evaluator*
            evaluator.add_event_handler(Events.COMPLETED, handler)

    .. versionchanged:: 0.6.0
        Renamed ``min_delta_mode``  to ``threshold_mode``.
        Renamed ``min_delta`` to ``threshold``.
        Renamed ``cumulative_delta`` to ``cumulative``.
    .. versionchanged:: 0.5.4
        Added `mode` parameter to support minimization in addition to maximization.
        Added `min_delta_mode` parameter to support both absolute and relative improvements.

    """

    _state_dict_all_req_keys = (
        "counter",
        "best_score",
        "threshold_mode",
    )

    def __init__(
        self,
        patience: int,
        score_function: Callable,
        trainer: Engine,
        threshold: float = 0.0,
        cumulative: bool = False,
        threshold_mode: Literal["abs", "rel"] = "abs",
        mode: Literal["min", "max"] = "max",
        # Deprecated args for BC
        min_delta: float | None = None,
        min_delta_mode: Literal["abs", "rel"] | None = None,
        cumulative_delta: bool | None = None,
    ):
        if not callable(score_function):
            raise TypeError("Argument score_function should be a function.")

        if patience < 1:
            raise ValueError("Argument patience should be positive integer.")

        if not isinstance(trainer, Engine):
            raise TypeError("Argument trainer should be an instance of Engine.")

        # Backward compatibility for deprecated args
        if min_delta is not None:
            warnings.warn(
                "'min_delta' is deprecated and will be removed in a future version. Please use 'threshold' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            threshold = min_delta

        if min_delta_mode is not None:
            warnings.warn(
                "'min_delta_mode' is deprecated and will be removed in a future version. "
                "Please use 'threshold_mode' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            threshold_mode = min_delta_mode

        if cumulative_delta is not None:
            warnings.warn(
                "'cumulative_delta' is deprecated and will be removed in a future version. "
                "Please use 'cumulative' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            cumulative = cumulative_delta

        if threshold < 0.0:
            raise ValueError("Argument threshold should not be a negative number.")

        if threshold_mode not in ("abs", "rel"):
            raise ValueError("Argument threshold_mode should be either 'abs' or 'rel'.")

        if mode not in ("min", "max"):
            raise ValueError("Argument mode should be either 'min' or 'max'.")

        self.score_function = score_function
        self.patience = patience
        self.threshold = threshold
        self.threshold_mode = threshold_mode
        self.cumulative = cumulative
        self.trainer = trainer
        self.counter = 0
        self.best_score: float | None = None
        self.logger = setup_logger(__name__ + "." + self.__class__.__name__)
        self.mode = mode

    @property
    def min_delta(self) -> float:
        warnings.warn(
            "min_delta is deprecated and will be removed in a future version. Please use 'threshold' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.threshold

    @min_delta.setter
    def min_delta(self, value: float) -> None:
        warnings.warn(
            "min_delta is deprecated and will be removed in a future version. Please use 'threshold' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.threshold = value

    @property
    def min_delta_mode(self) -> str:
        warnings.warn(
            "min_delta_mode is deprecated and will be removed in a future version. Please use 'threshold_mode' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.threshold_mode

    @min_delta_mode.setter
    def min_delta_mode(self, value: str) -> None:
        warnings.warn(
            "min_delta_mode is deprecated and will be removed in a future version. Please use 'threshold_mode' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.threshold_mode = value

    @property
    def cumulative_delta(self) -> bool:
        warnings.warn(
            "cumulative_delta is deprecated and will be removed in a future version. Please use 'cumulative' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.cumulative

    @cumulative_delta.setter
    def cumulative_delta(self, value: bool) -> None:
        warnings.warn(
            "cumulative_delta is deprecated and will be removed in a future version. Please use 'cumulative' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.cumulative = value

    def __call__(self, engine: Engine) -> None:
        score = self.score_function(engine)

        if self.best_score is None:
            self.best_score = score
            return

        threshold = -self.threshold if self.mode == "min" else self.threshold
        if self.threshold_mode == "abs":
            improvement_threshold = self.best_score + threshold
        else:
            improvement_threshold = self.best_score * (1 + threshold)

        no_improvement = score <= improvement_threshold if self.mode == "max" else score >= improvement_threshold

        if no_improvement:
            if not self.cumulative:
                self.best_score = max(score, self.best_score) if self.mode == "max" else min(score, self.best_score)
            self.counter += 1
            self.logger.debug("EarlyStopping: %i / %i" % (self.counter, self.patience))
            if self.counter >= self.patience:
                self.logger.info("EarlyStopping: Stop training")
                self.trainer.terminate()
        else:
            self.best_score = score
            self.counter = 0

    def reset(self) -> None:
        """Reset the early stopping state, including the counter and best score.

        .. versionadded:: 0.5.4
        """
        self.counter = 0
        self.best_score = None

    def attach(  # type: ignore[override]
        self,
        engine: Engine,
        event: Any = Events.COMPLETED,
        reset_engine: Engine | None = None,
        reset_event: Any = Events.STARTED,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Attaches the early stopping handler to an engine and registers its reset callback.

        This method will:
        1. Add the early stopping evaluation logic (``self``) to ``engine`` on the given ``event``.
        2. Add the ``reset`` method to ``reset_engine`` (or ``engine`` if not provided) on the given ``reset_event``.

        Args:
            engine: The engine to attach the early stopping evaluation to (typically an evaluator).
            event: The event on ``engine`` that triggers the early stopping check. Default is
                :attr:`~ignite.engine.events.Events.COMPLETED`.
            reset_engine: The engine to attach the reset callback to (typically the trainer).
                If ``None``, defaults to ``engine``.
            reset_event: The event on ``reset_engine`` that triggers the handler state reset.
                Default is :attr:`~ignite.engine.events.Events.STARTED`.

        .. versionadded:: 0.5.4
        """
        engine.add_event_handler(event, self)

        target_reset_engine = reset_engine or engine
        target_reset_engine.add_event_handler(reset_event, self.reset)

    def state_dict(self) -> "OrderedDict[str, Any]":
        """Method returns state dict with ``counter`` and ``best_score``.
        Can be used to save internal state of the class.
        """
        return OrderedDict(
            [
                ("counter", self.counter),
                ("best_score", cast(float, self.best_score)),
                ("threshold_mode", self.threshold_mode),
            ]
        )

    def load_state_dict(self, state_dict: Mapping) -> None:
        """Method replace internal state of the class with provided state dict data.

        Args:
            state_dict: a dict with "counter" and "best_score" keys/values.
        """
        super().load_state_dict(state_dict)
        self.counter = state_dict["counter"]
        self.best_score = state_dict["best_score"]
        self.threshold_mode = state_dict.get("threshold_mode", self.threshold_mode)



import warnings
from collections.abc import Mapping
from typing import Any

from torch.optim.optimizer import Optimizer

from ignite.engine.engine import Engine
from ignite.engine.events import Events
from ignite.handlers import (
    ClearMLLogger,
    MLflowLogger,
    NeptuneLogger,
    PolyaxonLogger,
    TensorboardLogger,
    VisdomLogger,
    WandBLogger,
    global_step_from_engine,
)
from ignite.handlers.base_logger import BaseLogger

__all__ = [
    "setup_tb_logging",
    "setup_visdom_logging",
    "setup_mlflow_logging",
    "setup_neptune_logging",
    "setup_wandb_logging",
    "setup_plx_logging",
    "setup_clearml_logging",
    "setup_trains_logging",
]

_MetricNames = str | list[str]


def _setup_logging(
    logger: BaseLogger,
    trainer: Engine,
    optimizers: Optimizer | dict[str, Optimizer] | dict[None, Optimizer] | None,
    evaluators: Engine | dict[str, Engine] | None,
    log_every_iters: int,
    *,
    trainer_metric_names: _MetricNames = "all",
    evaluator_metric_names: _MetricNames = "all",
) -> None:
    if optimizers is not None:
        if not isinstance(optimizers, (Optimizer, Mapping)):
            raise TypeError("Argument optimizers should be either a single optimizer or a dictionary or optimizers")

    if evaluators is not None:
        if not isinstance(evaluators, (Engine, Mapping)):
            raise TypeError("Argument evaluators should be either a single engine or a dictionary or engines")

    if log_every_iters is None:
        log_every_iters = 1

    logger.attach_output_handler(
        trainer,
        event_name=Events.ITERATION_COMPLETED(every=log_every_iters),
        tag="training",
        metric_names=trainer_metric_names,
    )

    if optimizers is not None:
        # Log optimizer parameters
        if isinstance(optimizers, Optimizer):
            optimizers = {None: optimizers}

        for k, optimizer in optimizers.items():
            logger.attach_opt_params_handler(
                trainer, Events.ITERATION_STARTED(every=log_every_iters), optimizer, param_name="lr", tag=k
            )

    if evaluators is not None:
        # Log evaluation metrics
        if isinstance(evaluators, Engine):
            evaluators = {"validation": evaluators}

        event_name = Events.ITERATION_COMPLETED if isinstance(logger, WandBLogger) else None
        gst = global_step_from_engine(trainer, custom_event_name=event_name)
        for k, evaluator in evaluators.items():
            logger.attach_output_handler(
                evaluator,
                event_name=Events.COMPLETED,
                tag=k,
                metric_names=evaluator_metric_names,
                global_step_transform=gst,
            )


def setup_tb_logging(
    output_path: str,
    trainer: Engine,
    optimizers: Optimizer | dict[str, Optimizer] | None = None,
    evaluators: Engine | dict[str, Engine] | None = None,
    log_every_iters: int = 100,
    *,
    trainer_metric_names: _MetricNames = "all",
    evaluator_metric_names: _MetricNames = "all",
    **kwargs: Any,
) -> TensorboardLogger:
    """Method to setup TensorBoard logging on trainer and a list of evaluators. Logged metrics are:

        - Training metrics, e.g. running average loss values
        - Learning rate(s)
        - Evaluation metrics

    Args:
        output_path: logging directory path
        trainer: trainer engine
        optimizers: single or dictionary of
            torch optimizers. If a dictionary, keys are used as tags arguments for logging.
        evaluators: single or dictionary of evaluators. If a dictionary,
            keys are used as tags arguments for logging.
        log_every_iters: interval for loggers attached to iteration events. To log every iteration,
            value can be set to 1 or None.
        trainer_metric_names: list of trainer metric names to plot or a string "all" to plot all available metrics.
        evaluator_metric_names: list of evaluator metric names to plot or a string "all" to plot all available metrics.
        kwargs: optional keyword args to be passed to construct the logger.

    .. versionchanged:: 0.6.0
        Added ``trainer_metric_names`` and ``evaluator_metric_names`` parameters.

    Returns:
        :class:`~ignite.handlers.tensorboard_logger.TensorboardLogger`

    Examples:
        .. code-block:: python

            from ignite.handlers.logger_utils import setup_tb_logging
            # Assume `trainer`, `evaluator`, and `optimizer` are already defined
            tb_logger = setup_tb_logging(
                output_path="experiments/tb_logs",
                trainer=trainer,
                optimizers=optimizer,
                evaluators={"validation": evaluator},
                log_every_iters=100
            )

            # Logger instance can be closed
            tb_logger.close()
    """
    logger = TensorboardLogger(log_dir=output_path, **kwargs)
    _setup_logging(
        logger,
        trainer,
        optimizers,
        evaluators,
        log_every_iters,
        trainer_metric_names=trainer_metric_names,
        evaluator_metric_names=evaluator_metric_names,
    )
    return logger


def setup_visdom_logging(
    trainer: Engine,
    optimizers: Optimizer | dict[str, Optimizer] | None = None,
    evaluators: Engine | dict[str, Engine] | None = None,
    log_every_iters: int = 100,
    *,
    trainer_metric_names: _MetricNames = "all",
    evaluator_metric_names: _MetricNames = "all",
    **kwargs: Any,
) -> VisdomLogger:
    """Method to setup Visdom logging on trainer and a list of evaluators. Logged metrics are:

        - Training metrics, e.g. running average loss values
        - Learning rate(s)
        - Evaluation metrics

    .. warning::

        This function uses VisdomLogger which is currently untested due to the visdom package being
        unmaintained and difficult to install with modern Python packages. Use at your own risk.

    Args:
        trainer: trainer engine
        optimizers: single or dictionary of
            torch optimizers. If a dictionary, keys are used as tags arguments for logging.
        evaluators: single or dictionary of evaluators. If a dictionary,
            keys are used as tags arguments for logging.
        log_every_iters: interval for loggers attached to iteration events. To log every iteration,
            value can be set to 1 or None.
        trainer_metric_names: list of trainer metric names to plot or a string "all" to plot all available metrics.
        evaluator_metric_names: list of evaluator metric names to plot or a string "all" to plot all available metrics.
        kwargs: optional keyword args to be passed to construct the logger.

    .. versionchanged:: 0.6.0
        Added ``trainer_metric_names`` and ``evaluator_metric_names`` parameters.

    Returns:
        :class:`~ignite.handlers.visdom_logger.VisdomLogger`

    Examples:
        .. code-block:: python

            from ignite.handlers.logger_utils import setup_visdom_logging
            # Assume `trainer`, `evaluator`, and `optimizer` are already defined
            vd_logger = setup_visdom_logging(
                trainer=trainer,
                optimizers=optimizer,
                evaluators={"validation": evaluator},
                log_every_iters=100
            )

            # Logger instance can be closed
            vd_logger.close()
    """
    logger = VisdomLogger(**kwargs)
    _setup_logging(
        logger,
        trainer,
        optimizers,
        evaluators,
        log_every_iters,
        trainer_metric_names=trainer_metric_names,
        evaluator_metric_names=evaluator_metric_names,
    )
    return logger


def setup_mlflow_logging(
    trainer: Engine,
    optimizers: Optimizer | dict[str, Optimizer] | None = None,
    evaluators: Engine | dict[str, Engine] | None = None,
    log_every_iters: int = 100,
    *,
    trainer_metric_names: _MetricNames = "all",
    evaluator_metric_names: _MetricNames = "all",
    **kwargs: Any,
) -> MLflowLogger:
    """Method to setup MLflow logging on trainer and a list of evaluators. Logged metrics are:

        - Training metrics, e.g. running average loss values
        - Learning rate(s)
        - Evaluation metrics

    Args:
        trainer: trainer engine
        optimizers: single or dictionary of
            torch optimizers. If a dictionary, keys are used as tags arguments for logging.
        evaluators: single or dictionary of evaluators. If a dictionary,
            keys are used as tags arguments for logging.
        log_every_iters: interval for loggers attached to iteration events. To log every iteration,
            value can be set to 1 or None.
        trainer_metric_names: list of trainer metric names to plot or a string "all" to plot all available metrics.
        evaluator_metric_names: list of evaluator metric names to plot or a string "all" to plot all available metrics.
        kwargs: optional keyword args to be passed to construct the logger.

    .. versionchanged:: 0.6.0
        Added ``trainer_metric_names`` and ``evaluator_metric_names`` parameters.

    Returns:
        :class:`~ignite.handlers.mlflow_logger.MLflowLogger`

    Examples:
        .. code-block:: python

            from ignite.handlers.logger_utils import setup_mlflow_logging
            # Assume `trainer`, `evaluator`, and `optimizer` are already defined
            mlflow_logger = setup_mlflow_logging(
                trainer=trainer,
                optimizers=optimizer,
                evaluators={"validation": evaluator},
                log_every_iters=100
            )

            # Logger instance can be closed
            mlflow_logger.close()
    """
    logger = MLflowLogger(**kwargs)
    _setup_logging(
        logger,
        trainer,
        optimizers,
        evaluators,
        log_every_iters,
        trainer_metric_names=trainer_metric_names,
        evaluator_metric_names=evaluator_metric_names,
    )
    return logger


def setup_neptune_logging(
    trainer: Engine,
    optimizers: Optimizer | dict[str, Optimizer] | None = None,
    evaluators: Engine | dict[str, Engine] | None = None,
    log_every_iters: int = 100,
    *,
    trainer_metric_names: _MetricNames = "all",
    evaluator_metric_names: _MetricNames = "all",
    **kwargs: Any,
) -> NeptuneLogger:
    """Method to setup Neptune logging on trainer and a list of evaluators. Logged metrics are:

        - Training metrics, e.g. running average loss values
        - Learning rate(s)
        - Evaluation metrics

    Args:
        trainer: trainer engine
        optimizers: single or dictionary of
            torch optimizers. If a dictionary, keys are used as tags arguments for logging.
        evaluators: single or dictionary of evaluators. If a dictionary,
            keys are used as tags arguments for logging.
        log_every_iters: interval for loggers attached to iteration events. To log every iteration,
            value can be set to 1 or None.
        trainer_metric_names: list of trainer metric names to plot or a string "all" to plot all available metrics.
        evaluator_metric_names: list of evaluator metric names to plot or a string "all" to plot all available metrics.
        kwargs: optional keyword args to be passed to construct the logger.

    .. versionchanged:: 0.6.0
        Added ``trainer_metric_names`` and ``evaluator_metric_names`` parameters.

    Returns:
        :class:`~ignite.handlers.neptune_logger.NeptuneLogger`

    Examples:
        .. code-block:: python

            from ignite.handlers.logger_utils import setup_neptune_logging
            # Assume `trainer`, `evaluator`, and `optimizer` are already defined
            neptune_logger = setup_neptune_logging(
                trainer=trainer,
                optimizers=optimizer,
                evaluators={"validation": evaluator},
                log_every_iters=100
            )

            # Logger instance can be closed
            neptune_logger.close()
    """
    logger = NeptuneLogger(**kwargs)
    _setup_logging(
        logger,
        trainer,
        optimizers,
        evaluators,
        log_every_iters,
        trainer_metric_names=trainer_metric_names,
        evaluator_metric_names=evaluator_metric_names,
    )
    return logger


def setup_wandb_logging(
    trainer: Engine,
    optimizers: Optimizer | dict[str, Optimizer] | None = None,
    evaluators: Engine | dict[str, Engine] | None = None,
    log_every_iters: int = 100,
    *,
    trainer_metric_names: _MetricNames = "all",
    evaluator_metric_names: _MetricNames = "all",
    **kwargs: Any,
) -> WandBLogger:
    """Method to setup WandB logging on trainer and a list of evaluators. Logged metrics are:

        - Training metrics, e.g. running average loss values
        - Learning rate(s)
        - Evaluation metrics

    Args:
        trainer: trainer engine
        optimizers: single or dictionary of
            torch optimizers. If a dictionary, keys are used as tags arguments for logging.
        evaluators: single or dictionary of evaluators. If a dictionary,
            keys are used as tags arguments for logging.
        log_every_iters: interval for loggers attached to iteration events. To log every iteration,
            value can be set to 1 or None.
        trainer_metric_names: list of trainer metric names to plot or a string "all" to plot all available metrics.
        evaluator_metric_names: list of evaluator metric names to plot or a string "all" to plot all available metrics.
        kwargs: optional keyword args to be passed to construct the logger.

    .. versionchanged:: 0.6.0
        Added ``trainer_metric_names`` and ``evaluator_metric_names`` parameters.

    Returns:
        :class:`~ignite.handlers.wandb_logger.WandBLogger`

    Examples:
        .. code-block:: python

            from ignite.handlers.logger_utils import setup_wandb_logging
            # Assume `trainer`, `evaluator`, and `optimizer` are already defined
            wandb_logger = setup_wandb_logging(
                trainer=trainer,
                optimizers=optimizer,
                evaluators={"validation": evaluator},
                log_every_iters=100
            )

            # Logger instance can be closed
            wandb_logger.close()
    """
    logger = WandBLogger(**kwargs)
    _setup_logging(
        logger,
        trainer,
        optimizers,
        evaluators,
        log_every_iters,
        trainer_metric_names=trainer_metric_names,
        evaluator_metric_names=evaluator_metric_names,
    )
    return logger


def setup_plx_logging(
    trainer: Engine,
    optimizers: Optimizer | dict[str, Optimizer] | None = None,
    evaluators: Engine | dict[str, Engine] | None = None,
    log_every_iters: int = 100,
    *,
    trainer_metric_names: _MetricNames = "all",
    evaluator_metric_names: _MetricNames = "all",
    **kwargs: Any,
) -> PolyaxonLogger:
    """Method to setup Polyaxon logging on trainer and a list of evaluators. Logged metrics are:

        - Training metrics, e.g. running average loss values
        - Learning rate(s)
        - Evaluation metrics

    Args:
        trainer: trainer engine
        optimizers: single or dictionary of
            torch optimizers. If a dictionary, keys are used as tags arguments for logging.
        evaluators: single or dictionary of evaluators. If a dictionary,
            keys are used as tags arguments for logging.
        log_every_iters: interval for loggers attached to iteration events. To log every iteration,
            value can be set to 1 or None.
        trainer_metric_names: list of trainer metric names to plot or a string "all" to plot all available metrics.
        evaluator_metric_names: list of evaluator metric names to plot or a string "all" to plot all available metrics.
        kwargs: optional keyword args to be passed to construct the logger.

    .. versionchanged:: 0.6.0
        Added ``trainer_metric_names`` and ``evaluator_metric_names`` parameters.

    Returns:
        :class:`~ignite.handlers.polyaxon_logger.PolyaxonLogger`

    Examples:
        .. code-block:: python

            from ignite.handlers.logger_utils import setup_plx_logging
            # Assume `trainer`, `evaluator`, and `optimizer` are already defined
            plx_logger = setup_plx_logging(
                trainer=trainer,
                optimizers=optimizer,
                evaluators={"validation": evaluator},
                log_every_iters=100
            )

            # Logger instance can be closed
            plx_logger.close()
    """
    logger = PolyaxonLogger(**kwargs)
    _setup_logging(
        logger,
        trainer,
        optimizers,
        evaluators,
        log_every_iters,
        trainer_metric_names=trainer_metric_names,
        evaluator_metric_names=evaluator_metric_names,
    )
    return logger


def setup_clearml_logging(
    trainer: Engine,
    optimizers: Optimizer | dict[str, Optimizer] | None = None,
    evaluators: Engine | dict[str, Engine] | None = None,
    log_every_iters: int = 100,
    *,
    trainer_metric_names: _MetricNames = "all",
    evaluator_metric_names: _MetricNames = "all",
    **kwargs: Any,
) -> ClearMLLogger:
    """Method to setup ClearML logging on trainer and a list of evaluators. Logged metrics are:

        - Training metrics, e.g. running average loss values
        - Learning rate(s)
        - Evaluation metrics

    Args:
        trainer: trainer engine
        optimizers: single or dictionary of
            torch optimizers. If a dictionary, keys are used as tags arguments for logging.
        evaluators: single or dictionary of evaluators. If a dictionary,
            keys are used as tags arguments for logging.
        log_every_iters: interval for loggers attached to iteration events. To log every iteration,
            value can be set to 1 or None.
        trainer_metric_names: list of trainer metric names to plot or a string "all" to plot all available metrics.
        evaluator_metric_names: list of evaluator metric names to plot or a string "all" to plot all available metrics.
        kwargs: optional keyword args to be passed to construct the logger.

    .. versionchanged:: 0.6.0
        Added ``trainer_metric_names`` and ``evaluator_metric_names`` parameters.

    Returns:
        :class:`~ignite.handlers.clearml_logger.ClearMLLogger`

    Examples:
        .. code-block:: python

            from ignite.handlers.logger_utils import setup_clearml_logging
            # Assume `trainer`, `evaluator`, and `optimizer` are already defined
            clearml_logger = setup_clearml_logging(
                trainer=trainer,
                optimizers=optimizer,
                evaluators={"validation": evaluator},
                log_every_iters=100
            )

            # Logger instance can be closed
            clearml_logger.close()
    """
    logger = ClearMLLogger(**kwargs)
    _setup_logging(
        logger,
        trainer,
        optimizers,
        evaluators,
        log_every_iters,
        trainer_metric_names=trainer_metric_names,
        evaluator_metric_names=evaluator_metric_names,
    )
    return logger


def setup_trains_logging(
    trainer: Engine,
    optimizers: Optimizer | dict[str, Optimizer] | None = None,
    evaluators: Engine | dict[str, Engine] | None = None,
    log_every_iters: int = 100,
    *,
    trainer_metric_names: _MetricNames = "all",
    evaluator_metric_names: _MetricNames = "all",
    **kwargs: Any,
) -> ClearMLLogger:
    """``setup_trains_logging`` was renamed to :func:`~ignite.handlers.logger_utils.setup_clearml_logging`.

    .. versionchanged:: 0.6.0
        Added ``trainer_metric_names`` and ``evaluator_metric_names`` parameters.
    """
    warnings.warn("setup_trains_logging was renamed to setup_clearml_logging.")
    return setup_clearml_logging(
        trainer,
        optimizers,
        evaluators,
        log_every_iters,
        trainer_metric_names=trainer_metric_names,
        evaluator_metric_names=evaluator_metric_names,
        **kwargs,
    )
